// $Id: SignaturePanel.java 16027 2012-07-13 11:18:01Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.*;
import javax.swing.*;
import java.awt.event.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.awt.*;
import java.text.DateFormat;
import java.util.*;
import java.util.List;

/**
 * Create a {@link SidePanel} that will display a list of Digital Signatures in the PDF.
 * Double clicking on a blank signature will select a {@link SignatureProvider} and
 * then call {@link SignatureProvider#showSignDialog SignatureProvider.showSignDialog()}.
 * Double clicking on a previously signed signature will select a {@link SignatureProvider} and
 * then call {@link SignatureProvider#showVerifyDialog SignatureProvider.showVerifyDialog()}.
 * <br/><br/>
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">ShowHideSignatures</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8
 */
public class SignaturePanel extends SidePanelFactory {


    /**
     * Return the SignaturePanel
     * @deprecated since 2.11 just calls the public constructor
     */
    public static SignaturePanel getInstance() {
        return new SignaturePanel();
    }

    /**
     * Create a new SignaturePanel
     * @since 2.11
     */
    public SignaturePanel() {
        super("ShowHideSignatures");
    }

    public boolean isSidePanelRequired(PDF pdf) {
        for (Iterator i = pdf.getForm().getElements().values().iterator();i.hasNext();) {
            if (i.next() instanceof FormSignature) {
                return true;
            }
        }
        return false;
    }

    public SidePanel createSidePanel() {
        return new SignaturePanelImpl(this);
    }
}

/**
 * A {@link SidePanel} representing the "Signature" or "Bookmark" panel.
 *
 * <p><i>
 * This code is copyright the Big Faceless Organization. You're welcome to
 * use, modify and distribute it in any form in your own projects, provided
 * those projects continue to make use of the Big Faceless PDF library.
 * </i></p>
 */
class SignaturePanelImpl extends JPanel implements SidePanel, ListCellRenderer
{
    private JScrollPane scrollpane;
    private DocumentPanel docpanel;
    private Icon icon;
    private SignaturePanel factory;

    SignaturePanelImpl(SignaturePanel factory) {
        super(new BorderLayout());
        setOpaque(true);
        this.factory = factory;
        scrollpane = new JScrollPane();
        add(scrollpane, BorderLayout.CENTER);
        this.icon = new ImageIcon(PDFViewer.class.getResource("resources/icons/key.png"));
    }

    public Icon getIcon() {
        return icon;
    }

    public String getName() {
        return "Signatures";
    }

    public void panelVisible() {
    }

    public void panelHidden() {
    }

    public void setDocumentPanel(final DocumentPanel docpanel) {
        this.docpanel = docpanel;
        final PDF pdf = docpanel==null ? null : docpanel.getPDF();
        final Vector v = new Vector();
        if (pdf!=null) {
            for (Iterator i = pdf.getForm().getElements().entrySet().iterator();i.hasNext();) {
                Map.Entry e = (Map.Entry)i.next();
                if (e.getValue() instanceof FormSignature) {
                    FormSignature sig = (FormSignature)e.getValue();
                    v.add(sig);
                }
            }
        }

        if (v.size()>0) {
            final JList list = new JList(v);
            list.setCellRenderer(this);
            list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
            list.addMouseListener(new MouseAdapter() {
                public void mouseClicked(MouseEvent event) {
                    if (event.getClickCount() == 2) {
                        int ix = list.locationToIndex(event.getPoint());
                        if (ix>=0) {
                            final FormSignature field = (FormSignature)v.get(ix);
                            if (field.getState()==FormSignature.STATE_SIGNED) {
                                SignatureProvider.selectVerifyProvider(docpanel, field, list, event.getPoint(), new ActionListener() {
                                    public void actionPerformed(ActionEvent event) {
                                        SignatureProvider ssp = (SignatureProvider)event.getSource();
                                        ssp.showVerifyDialog(docpanel, field);
                                    }
                                });
                            } else {
                                SignatureProvider.selectSignProvider(docpanel, field, list, event.getPoint(), new ActionListener() {
                                    public void actionPerformed(ActionEvent event) {
                                        SignatureProvider ssp = (SignatureProvider)event.getSource();
                                        try {
                                            ssp.showSignDialog(docpanel, field);
                                        } catch (Exception e) {
                                            Util.displayThrowable(e, docpanel);
                                        }
                                    }
                                });
                            }
                        }
                    }
                }
            });

            JPanel panel = new JPanel(new GridBagLayout());
            panel.setBackground(UIManager.getColor("List.background"));
            panel.setOpaque(true);
            GridBagConstraints gbc = new GridBagConstraints();
            gbc.fill = gbc.BOTH;
            gbc.weightx = 1;
            gbc.anchor = gbc.WEST;
            gbc.gridwidth = gbc.REMAINDER;
            panel.add(list, gbc);
            JButton verifyall = new JButton("VerifyAll");
            verifyall.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    new BackgroundSignatureVerifier().verify(docpanel, v);
                    list.repaint();
                }
            });
            gbc.fill = gbc.NONE;
            panel.add(verifyall, gbc);
            
            if (factory.getFeatureProperty(docpanel.getViewer(), "VerifyAll.Hide") != null) {
                verifyall.setVisible(false);
            }

            gbc.fill = gbc.BOTH;
            gbc.weighty = 1;
            JPanel fill = new JPanel();
            fill.setOpaque(false);
            panel.add(fill, gbc);

            scrollpane.setVisible(true);
            scrollpane.setViewportView(panel);
        } else {
            scrollpane.setVisible(false);
            scrollpane.setViewportView(null);
        }
        repaint();
    }

    public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean hasFocus) {
        FormSignature field = (FormSignature)value;
        String fieldname = field.getDescription();
        if (fieldname==null) {
            fieldname = field.getForm().getName(field);
        }
        Icon icon;
        StringBuffer sb = new StringBuffer();
        sb.append("<b>"+fieldname+"</b><br>");
        if (field.getState()==FormSignature.STATE_SIGNED) {
            icon = SignatureProvider.getIcon(docpanel, field);
            if (field.getSignDate()!=null) {
                sb.append(UIManager.getString("PDFViewer.Date")+": ");
                Date d = field.getSignDate().getTime();
                DateFormat df = DateFormat.getDateTimeInstance();
                sb.append(df.format(d)+"<br>");
            }
            if (field.getName()!=null) {
                sb.append(UIManager.getString("PDFViewer.Name")+": ");
                sb.append(field.getName()+"<br>");
            }
            if (field.getReason()!=null) {
                sb.append(UIManager.getString("PDFViewer.Reason")+": ");
                sb.append(field.getReason()+"<br>");
            }
            if (field.getLocation()!=null) {
                sb.append(UIManager.getString("PDFViewer.Location")+": ");
                sb.append(field.getLocation()+"<br>");
            }
        } else {
            icon = new ImageIcon(PDFViewer.class.getResource("resources/icons/singlepage.png"));
        }

        JLabel comp = new JLabel("<html>"+sb+"</html>");
        comp.setBorder(BorderFactory.createEmptyBorder(0, 0, 10, 0));
        comp.setIcon(icon);
        comp.setOpaque(true);
        comp.setBackground(isSelected ? list.getSelectionBackground() : list.getBackground());
        comp.setForeground(isSelected ? list.getSelectionForeground() : list.getForeground());
        comp.setVerticalTextPosition(SwingConstants.TOP);
        return comp;
    }
}
